/*
    Valtra Tractors
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <algorithm>
#include "sprites.h"

using std::min;
using std::max;

bool operator==(const SDL_Rect &a, const SDL_Rect &b) {
	if (a.x!=b.x) return false;
	if (a.y!=b.y) return false;
	if (a.w!=b.w) return false;
	if (a.h!=b.h) return false;
	return true;
}

Sprite::Sprite(SDL_Surface *s, const SDL_Rect &p, bool u) {
	img = s;
	pos = p;
	update = u;
	old = false;
	relative = false;
	multUpdate = false;
	updateRect.x = 0;
	updateRect.y = 0;
	if (update) {
		if (img) {
			updateRect.w = img->w;
			updateRect.h = img->h;
		} else {
			updateRect.w = pos.w;
			updateRect.h = pos.h;
		}
	} else {
		updateRect.w = 0;
		updateRect.h = 0;
	}
	if (img) {
		pos.w = img->w;
		pos.h = img->h;
	}
}

Sprite::Sprite(SDL_Surface *s, const SDL_Rect &p, const SDL_Rect &u) {
	img = s;
	pos = p;
	update = true;
	old = false;
	relative = false;
	multUpdate = false;
	updateRect = u;
	if (img) {
		pos.w = img->w;
		pos.h = img->h;
	}
}

Sprite::Sprite(SDL_Surface *s, const SDL_Rect &p) {
	img = s;
	pos = p;
	update = true;
	old = true;
	relative = false;
	multUpdate = true;
	pos.w = img->w;
	pos.h = img->h;
}


SpriteEngine::SpriteEngine() {
	nextId = 1;
	curSprites = NULL;
	lastSprites = NULL;
	updatedRect = NULL;
}

void SpriteEngine::deleteVector(vector<Sprite*> *vect) {
	if (vect) {
		for (unsigned int i = 0; i < vect->size(); i++)
			delete (*vect)[i];
		delete vect;
	}
}

SpriteEngine::~SpriteEngine() {
	deleteVector(curSprites);
	deleteVector(lastSprites);
}

int SpriteEngine::getNextId() {
	return nextId++;
}

void SpriteEngine::addSprite(int id, Sprite *sprite) {
	if (curSprites==NULL) {
		curSprites = new vector<Sprite*>();
		if (nextId>1 && lastSprites!=NULL) {
			if (nextId>int(lastSprites->size())) {
				curSprites->reserve(nextId);
			} else {
				curSprites->reserve(lastSprites->size());
			}
		}
	}
	sprite->id = id;
	curSprites->push_back(sprite);
}

void SpriteEngine::redrawRect(const SDL_Rect &r) {
	rects.push_back(r);
}

void SpriteEngine::redrawAll() {
	deleteVector(lastSprites);
	lastSprites = NULL;
	rects.clear();
}

bool SpriteEngine::rectInside(const SDL_Rect &a, const SDL_Rect &b) {
	if (a.x<b.x) return false;
	if (a.y<b.y) return false;
	if (a.x+a.w>b.x+b.w) return false;
	if (a.y+a.h>b.y+b.h) return false;
	return true;
}

bool SpriteEngine::rectsOverlapMuch(const SDL_Rect &a, const SDL_Rect &b) {
	int area1 = a.w*a.h;
	int area2 = b.w*b.h;
	SDL_Rect out;
	if (!rectsIntersect(a,b,out)) return false;
	int area3 = out.w*out.h;
	if (area3>area1/10 || area3>area2/10) return true;
	return false;
}

bool SpriteEngine::rectsIntersect(const SDL_Rect &a, const SDL_Rect &b, SDL_Rect &out) {
	out.x = max(a.x,b.x);
	out.y = max(a.y,b.y);
	int w = min(a.x+a.w,b.x+b.w)-out.x; // SDL_Rect.w,h are unsigned!
	int h = min(a.y+a.h,b.y+b.h)-out.y;
	if (w>0 && h>0) {
		out.w = w;
		out.h = h;
		return true;
	}
	return false;
}

void SpriteEngine::mergeRects(const SDL_Rect &a, const SDL_Rect &b, SDL_Rect &out) {
	if (a.w==0 || a.h==0) {
		out = b;
		return;
	} else if (b.w==0 || b.h==0) {
		out = a;
		return;
	}
	out.x = min(a.x,b.x);
	out.y = min(a.y,b.y);
	out.w = max(a.x+a.w,b.x+b.w)-out.x;
	out.h = max(a.y+a.h,b.y+b.h)-out.y;
}

void SpriteEngine::calcUpdates() {
	if (curSprites==NULL) curSprites = new vector<Sprite*>();

	if (lastSprites==NULL) {
		if (curSprites==NULL) return;
		for (unsigned int i = 0; i < curSprites->size(); i++) {
			Sprite *s = (*curSprites)[i];
			s->update = true;
/*			s->updateRect.x = 0;
			s->updateRect.y = 0;
			s->updateRect.w = s->pos.w;
			s->updateRect.h = s->pos.h;
*/		}
		return;
	}

	for (unsigned int i = 0; i < lastSprites->size(); i++) {
		Sprite *s1 = (*lastSprites)[i];
		if (s1->id==0) {
			rects.push_back(s1->pos);
			continue;
		}
		bool found = false;
		for (unsigned int j = 0; j < curSprites->size(); j++) {
			Sprite *s2 = (*curSprites)[j];
			if (s1->id==s2->id) {
				if (s1->pos==s2->pos && s1->multUpdate==s2->multUpdate) {
					s2->old = true;
					found = true;
				}
			}
		}
		if (!found) rects.push_back(s1->pos);
	}

	for (unsigned int i = 0; i < curSprites->size(); i++) {
		Sprite *s = (*curSprites)[i];
		if (s->multUpdate) continue;
		if (!s->old) {
			s->update = true;
			s->updateRect.x = 0;
			s->updateRect.y = 0;
			s->updateRect.w = s->pos.w;
			s->updateRect.h = s->pos.h;
		}
		if (s->update) rects.push_back(s->pos);
	}

	removeInsideRects();

	for (unsigned int i = 0; i < curSprites->size(); i++) {
		Sprite *s = (*curSprites)[i];
		if (s->multUpdate) continue;
		for (unsigned int j = 0; j < rects.size(); j++) {
			SDL_Rect out;
			if (rectsIntersect(s->pos,rects[j],out)) {
				out.x -= s->pos.x;
				out.y -= s->pos.y;
				if (s->update) {
					SDL_Rect newRect;
					mergeRects(out,s->updateRect,newRect); // because of this, the background should be managed separately
					s->updateRect = newRect;
				} else {
					s->updateRect = out;
					s->update = true;
				}
			}
		}
	}

}

void SpriteEngine::removeInsideRects() {
	for (unsigned int i = 0; i < rects.size(); i++) {
		for (unsigned int j = 0; j < rects.size(); j++) {
			if (j==i) continue;
			if (rectInside(rects[j],rects[i])) {
				rects.erase(rects.begin()+j);
				if (j<i) i--;
				j--;
			}
		}
	}
}

void SpriteEngine::mergeRects(vector<SDL_Rect> &list, bool all) {
	for (unsigned int i = 0; i < list.size(); i++) {
		for (unsigned int j = 0; j < list.size(); j++) {
			if (j==i) continue;
			bool merge = false;
			if (all) {
				SDL_Rect r;
				merge = rectsIntersect(list[j],list[i],r);
			} else {
				merge = rectsOverlapMuch(list[j],list[i]);
			}
			if (merge) {
				SDL_Rect newRect;
				mergeRects(list[i],list[j],newRect);
				list[i] = newRect;
				list.erase(list.begin()+j);
				if (j<i) i--;
				j--;
				i--; // recheck this rect with all other
				break;
			}
		}
	}
}

void SpriteEngine::mergeRects(bool all) {
	mergeRects(rects,all);
}

int SpriteEngine::countSides(const SDL_Rect &a, const SDL_Rect &b) {
	int n = 0;
	if (a.x==b.x) n++;
	if (a.y==b.y) n++;
	if (a.x+a.w==b.x+b.w) n++;
	if (a.y+a.h==b.y+b.h) n++;
	return n;
}

void SpriteEngine::splitOverlap(vector<SDL_Rect> &list) {
	for (unsigned int i = 0; i < list.size(); i++) {
		for (unsigned int j = i+1; j < list.size(); j++) {
			if (j==i) continue;
			SDL_Rect r;
			if (!rectsIntersect(list[j],list[i],r)) continue;

			int sides1 = countSides(r,list[i]);
			int sides2 = countSides(r,list[j]);

//			if (rectInside(list[j],list[i])) {
			if (sides1==4) {
				// assert(i<j);
				list[i] = list[j];
				list.erase(list.begin()+j);
				j--;
				continue;
//			} else if (rectInside(list[i],list[j])) {
			} else if (sides2==4) {
				// assert(i<j);
				list.erase(list.begin()+j);
				j--;
				continue;
			}

			SDL_Rect split, out;
			if (sides1>sides2) {
				split = list[i];
				list[i] = list[j];
			} else {
				split = list[j];
			}
			list.erase(list.begin()+j);
			j--;

			while (true) {
				if (split.y==r.y && split.y+split.h>r.y+r.h) {
					// split loose a row from the bottom
					out.y = r.y+r.h;
					out.h = split.y+split.h-out.y;
					out.x = split.x;
					out.w = split.w;
					list.push_back(out);
					split.h = out.y-split.y;
					continue;
				}
				if (split.y<r.y && split.y+split.h==r.y+r.h) {
					// split loose a row from the top
					out.y = split.y;
					out.h = r.y-split.y;
					out.x = split.x;
					out.w = split.w;
					list.push_back(out);
					split.y = r.y;
					split.h = r.h;
					continue;
				}
				if (split.x==r.x && split.x+split.w>r.x+r.w) {
					// split loose a column from the right
					out.x = r.x+r.w;
					out.w = split.x+split.w-out.x;
					out.y = split.y;
					out.h = split.h;
					list.push_back(out);
					split.w = out.x-split.x;
					continue;
				}
				if (split.x<r.x && split.x+split.w==r.x+r.w) {
					// split loose a column from the left
					out.x = split.x;
					out.w = r.x-split.x;
					out.y = split.y;
					out.h = split.h;
					list.push_back(out);
					split.x = r.x;
					split.w = r.w;
					continue;
				}
				if (split.x==r.x && split.x+split.w==r.x+r.w) {
					// split into two rows
					out.x = split.x;
					out.y = split.y;
					out.w = split.w;
					out.h = r.y-split.y;
					if (out.h>0)
						list.push_back(out);
					out.y = r.y+r.h;
					out.h = split.y+split.h-out.y;
					if (out.h>0)
						list.push_back(out);
					break;
				}
				if (split.y==r.y && split.y+split.h==r.y+r.h) {
					// split into two columns
					out.x = split.x;
					out.y = split.y;
					out.w = r.x-split.x;
					out.h = split.h;
					if (out.w>0)
						list.push_back(out);
					out.x = r.x+r.w;
					out.w = split.x+split.w-out.x;
					if (out.w>0)
						list.push_back(out);
					break;
				}
			}

		}
	}
}


void SpriteEngine::update(SDL_Surface *screen, SDL_Surface *bkg) {
	if (lastSprites==NULL) {
		if (bkg!=NULL)
			SDL_BlitSurface(bkg,NULL,screen,NULL);
		for (unsigned int i = 0; i < curSprites->size(); i++) {
			Sprite *s = (*curSprites)[i];
			SDL_BlitSurface(s->img,NULL,screen,&s->pos);
		}
		SDL_Rect r;
		r.x = 0;
		r.y = 0;
		r.w = screen->w;
		r.h = screen->h;
		if (updatedRect)
			updatedRect(r);
		else
			rects.push_back(r);
	} else {
		if (bkg!=NULL) {
			for (unsigned int i = 0; i < rects.size(); i++)
				SDL_BlitSurface(bkg,&rects[i],screen,&rects[i]);
		}

		vector<SDL_Rect> merged = rects;
//		mergeRects(merged,true);
		splitOverlap(merged);

		for (unsigned int i = 0; i < curSprites->size(); i++) {
			Sprite *s = (*curSprites)[i];
			if (s->multUpdate) {
				for (unsigned int j = 0; j < merged.size(); j++) {
					SDL_Rect out;
					if (rectsIntersect(s->pos,merged[j],out)) {
						SDL_Rect srcPos = out;
						srcPos.x -= s->pos.x;
						srcPos.y -= s->pos.y;
						SDL_BlitSurface(s->img,&srcPos,screen,&out);
					}
				}
			} else {
				if (s->update) {
					SDL_Rect out = s->pos;
					if (s->updateRect.x>0 || s->updateRect.y>0) {
						out.x += s->updateRect.x;
						out.y += s->updateRect.y;
					}
					if (s->relative) {
						s->updateRect.x = 0;
						s->updateRect.y = 0;
					}
					SDL_BlitSurface(s->img,&s->updateRect,screen,&out);
				}
			}
		}

//		mergeRects();

		if (updatedRect) {
			for (unsigned int i = 0; i < rects.size(); i++)
				updatedRect(rects[i]);
		}
	}
}

void SpriteEngine::reset() {
	deleteVector(lastSprites);
	lastSprites = curSprites;
	curSprites = NULL;
	nextId = 1;
	lastRects = rects;
	rects.clear();
}

void SpriteEngine::setUpdatedRect(void (*func)(const SDL_Rect &r)) {
	updatedRect = func;
}

Sprite *SpriteEngine::getSprite(int id) {
	if (id==0) return NULL;
	for (unsigned int i = 0; i < curSprites->size(); i++) {
		Sprite *s = (*curSprites)[i];
		if (s->id==id) return s;
	}
	return NULL;
}

void SpriteEngine::getRects(vector<SDL_Rect> &list) {
	list.clear();
	for (unsigned int i = 0; i < lastRects.size(); i++)
		list.push_back(rects[i]);
}

